# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                       (C)2015-16, 2019-20 Jeremy Brown                       #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from os import listdir, remove
from os.path import isfile, join

from apic_tool.workers import get_format_worker


def get_music_files(logger, files, dirs, forced):
	"""
	Obtain a list of all music files among the provided files and directories
	that the tool is capable of adding images to.

	:param logger: (Logger) Logging object
	:param files: (list) Strings representing absolute paths to music files
	:param dirs: (list) Strings representing absulte paths to directories
						containing music files
	:param forced: (bool) Whether or not the tool should allow things to happen
						  that may have complications

	:returns: (list) Strings representing absolute paths to manipulable music files
	"""
	valid_files = []
	paths = []

	if files is not None:
		paths.extend(files)

	if dirs is not None:
		for dirfiles in [list(filter(isfile, map(lambda f: join(d, f), listdir(d)))) for d in dirs]:
			paths.extend(dirfiles)

	for path in paths:
		worker = get_format_worker(path)

		if worker is None:
			logger.debug("File %s is not a supported music file, skipping", path)
			continue

		if worker.can_insert_image(logger, path, forced):
			valid_files.append(path)

	return valid_files


def insert_image(logger, cover_path, insertion_dirs, insertion_files, keep_cover, dry_run, forced):
	"""
	Dispatch function handling qualifying files to insert images into
	and actually performing insertion.

	:param logger: (Logger) Logging object
	:param cover_path: (str) Absolute path to cover to insert into music files
	:param insertion_dirs: (list) Strings representing absolute paths to dirs
								  possibly containing music files
	:param insertion_files: (list) Strings representing absolute paths to
								   music files for possible insertion
	:param keep_cover: (bool) Whether or not to keep the inserted cover in the
							  event of success
	:param dry_run: (bool) Whether or not to perform the actual insertion of
						   images into files and deletion of cover afterwards
	:param forced: (bool) Whether or not the tool should allow things to happen
						  that may have complications
	"""
	result = True

	music_files = get_music_files(logger, insertion_files, insertion_dirs, forced)

	if music_files:
		for track in music_files:
			logger.debug("Writing image %s to file %s", cover_path, track)
			if not dry_run:
				worker = get_format_worker(track)
				result &= worker.write_to_metadata(logger, track, cover_path, forced)

		if result and not keep_cover:
			logger.info("Deleting image file %s", cover_path)
			if not dry_run:
				remove(cover_path)



# encoding: utf-8

################################################################################
#                                 id3autosort                                  #
#                      Sort audio files based on metadata                      #
#                    (C)2009-10, 2015, 2019-20 Jeremy Brown                    #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

import re

from argparse import Action, ArgumentParser, ArgumentTypeError
from logging import (
	DEBUG,
	ERROR,
	Formatter,
	getLogger,
	INFO,
	StreamHandler,
	WARNING,
	)
from os import access, makedirs, walk, R_OK, sep, W_OK
from os.path import abspath, expanduser, isdir, join
from sys import argv

from id3autosort import __version__
from id3autosort.sorter import sort


logger = getLogger(__file__)


class CustomLogs(Formatter):
	FORMATS = {
		ERROR:		"[-] %(message)s",
		WARNING:	"[~] %(message)s",
		INFO:		"[*] %(message)s",
		DEBUG:		"[.] %(message)s",
		}

	def __init__(self):
		super(CustomLogs, self).__init__()

	def format(self, record):
		log_style = getattr(self, "_style", self)
		log_style._fmt = self.FORMATS.get(record.levelno, self.FORMATS[DEBUG])
		return super(CustomLogs, self).format(record)


def parse_args(**kwargs):
	"""
	Read arguments from stdin while validating and performing any necessary conversions

	:returns: (Namespace) Tool arguments
	"""
	def _absolute_writable_path(path):
		expanded_path = abspath(expanduser(path))

		if not isdir(expanded_path):
			raise ArgumentTypeError("The given path is not a directory: {0}".format(expanded_path))

		if not access(expanded_path, R_OK | W_OK):
			raise ArgumentTypeError("User cannot read/write to the directory: {0}".format(expanded_path))

		return expanded_path


	def _directory_structure(raw_structure):
		expanded_structure = []
		subs = {"r": "{artist}", "l": "{album}", "d": "{date}", "g": "{genre}"}
		split_structure = [level for level in raw_structure.split(sep) if level != ""]

		for level in split_structure:
			pattern = re.compile("|".join(subs.keys()))
			expanded_structure.append(pattern.sub(lambda x: subs[x.group(0)], level))

		return sep.join(expanded_structure)


	parser = ArgumentParser(
		prog = "id3autosort",
		epilog = "(C) 2009-10, 2015, 2019-20 Jeremy Brown; Released under Non-Profit Open Source License version 3.0",
		description = "Organize music libraries based on each track's metadata."
		)

	parser.add_argument("src_paths",
						type=_absolute_writable_path,
						metavar="input_path",
						nargs="+",
						help="Directory containing audio files to organize"
						)

	parser.add_argument("dest_path",
						type=_absolute_writable_path,
						metavar="output_path",
						help="Directory audio files should be sorted into"
						)

	parser.add_argument("-s", "--structure",
						type=_directory_structure,
						default=sep.join(["r", "l"]),
						help = "Specify structure used to organize sorted MP3s")

	parser.add_argument("-u", "--windows-unsafe",
						dest="windows_safe",
						action="store_false",
						help=("Use all characters in metadata for new directories, "
							  "including ones Windows filesystems normally choke on")
						)

	parser.add_argument("-v", "--verbose",
						action="store_true",
						help="Increase logging verbosity")

	parser.add_argument("-n", "--dry-run",
						action="store_true",
						help="Don't actually move music files"
						)

	parser.add_argument("--version",
						action="version",
						version="%(prog)s {}".format(__version__))

	args = parser.parse_args(kwargs.get("argv", argv[1:]))
	
	return args


def main():
	"""
	Tool entry point
	"""
	args = parse_args()

	logger.setLevel(DEBUG if args.verbose else INFO)
	log_hdlr = StreamHandler()
	log_hdlr.setFormatter(CustomLogs())
	logger.addHandler(log_hdlr)

	logger.debug("Dry run: %s", args.dry_run)
	for path in args.src_paths:
		logger.debug("Source path: %s", path)
	logger.debug("Destination structure: %s%s%s", args.dest_path, sep, args.structure)
	logger.debug("Windows-safe directories: %s", args.windows_safe)

	for path in args.src_paths:
		sort(logger, path, args.dest_path, args.structure, args.windows_safe, args.dry_run)



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                           (C) 2015-16, 2019 Mischif                          #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from argparse import ArgumentTypeError, Namespace
from mock import Mock, patch
from os.path import abspath, dirname, getsize, join

import pytest

from apic_tool.cli import main, parse_args


APIC_TOOL_DATA = abspath(join(dirname(__file__), "data"))
AUTOSORT_AUDIO = abspath(join(dirname(dirname(__file__)), "id3autosort", "audio"))


@pytest.mark.parametrize("action", ["insert", "extract"], ids=["insert-action", "extract-action"])
@pytest.mark.parametrize("dry_run", [True, False], ids=["dry-run", "no-dry-run"])
@pytest.mark.parametrize("verbose", [True, False], ids=["verbose", "no-verbose"])
@pytest.mark.parametrize("force", [True, False], ids=["forced", "not-forced"])
@pytest.mark.parametrize("extract_pic", [True, False], ids=["extract-location", "no-extract-location"])
@pytest.mark.parametrize("keep_pic", [True, False], ids=["keep-pic", "no-keep-pic"])
@pytest.mark.parametrize("insert_dirs, insert_files", [(True, False), (False, True)], ids=["insert-dirs", "insert-files"])
def test_parse_args(action, dry_run, verbose, force, extract_pic, keep_pic, insert_dirs, insert_files, tmpdir):
	extract_location = None
	args_list = []

	if dry_run:
		args_list.append("-n")

	if verbose:
		args_list.append("-v")

	if force:
		args_list.append("--force")

	args_list.append(action)

	if action == "extract":
		args_list.append(join(APIC_TOOL_DATA, "test_extract.mp3"))

		if extract_pic:
			extract_location = join(APIC_TOOL_DATA, "test_extract.png")
			args_list.append(extract_location)

	elif action == "insert":
		args_list.extend(["-p", join(APIC_TOOL_DATA, "test_cover.png")])

		if keep_pic:
			args_list.append("-k")

		if insert_files:
			args_list.extend(["-f", join(APIC_TOOL_DATA, "test_insert.mp3"),
							  "--file", join(APIC_TOOL_DATA, "test_extract.mp3")])

		if insert_dirs:
			args_list.extend(["-d", APIC_TOOL_DATA])

	args = parse_args(argv=args_list)

	assert args.dry_run == dry_run
	assert args.verbose == verbose
	assert args.force == force

	if action == "extract":
		assert args.extract_music == join(APIC_TOOL_DATA, "test_extract.mp3")
		assert args.extract_pic == extract_location

	if action == "insert":
		assert args.keep_pic == keep_pic

		if insert_files:
			assert args.insert_files == [join(APIC_TOOL_DATA, "test_insert.mp3"),
										 join(APIC_TOOL_DATA, "test_extract.mp3")]

		if insert_dirs:
			assert args.insert_dirs == [APIC_TOOL_DATA]


def test_action_unhappy_paths(tmpdir):
	with pytest.raises(ArgumentTypeError):
		parse_args(argv=["extract", str(tmpdir)])

	with pytest.raises(ArgumentTypeError):
		parse_args(argv=["extract", join(AUTOSORT_AUDIO, "test_wav.wav")])

	with pytest.raises(ArgumentTypeError):
		parse_args(argv=["insert", "-d", join(AUTOSORT_AUDIO, "test_wav.wav", "-p", join(APIC_TOOL_DATA, "test_cover.png"))])

	with pytest.raises(ArgumentTypeError):
		parse_args(argv=["insert", "-f", join(APIC_TOOL_DATA, "test_insert.mp3"), "-p", join(APIC_TOOL_DATA, "test_unsupported_cover.bmp")])


@pytest.mark.parametrize("action", ["insert", "extract"], ids=["insert-action", "extract-action"])
@patch("apic_tool.cli.insert_image")
@patch("apic_tool.cli.extract_image")
@patch("apic_tool.cli.parse_args")
@patch("apic_tool.cli.logger")
def test_main(mock_logger, mock_parse_args, mock_extract, mock_insert, action):
	args_dict = {
		"action": action,
		"dry_run": False,
		"extract_music": join(APIC_TOOL_DATA, "test_extract.mp3"),
		"extract_pic": None,
		"force": False,
		"insert_files": [join(APIC_TOOL_DATA, "test_extract.mp3")],
		"insert_dirs": None,
		"insert_pic": join(APIC_TOOL_DATA, "test_cover.png"),
		"keep_pic": True,
		"verbose": False,
		}

	mock_parse_args.return_value = Namespace(**args_dict)
	main()

	if action == "extract":
		mock_extract.assert_called_once_with(mock_logger,
											 args_dict["extract_music"],
											 args_dict["extract_pic"],
											 args_dict["dry_run"],
											 args_dict["force"])
	else:
		mock_insert.assert_called_once_with(mock_logger,
											args_dict["insert_pic"],
											args_dict["insert_dirs"],
											args_dict["insert_files"],
											args_dict["keep_pic"],
											args_dict["dry_run"],
											args_dict["force"])



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                       (C)2015-16, 2019-20 Jeremy Brown                       #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from errno import EACCES
from os.path import isfile

from apic_tool.workers import get_format_worker


def write_to_disk(logger, path, data):
	"""
	Write the given data to the given location.

	:param logger: (Logger) Logging object
	:param path: (str) Absolute path to write data to
	:param data: (bytes) Data to write to a file
	"""
	try:
		with open(path, "wb") as image:
			image.write(data)
	except IOError as e:
		if e.errno == EACCES:
			logger.info("Permission denied attempting to write image %s", path)
		else:
			raise


def fuzzy_match(user_ext, worker_ext):
	"""
	Check if the extensions the user provided and the worker determined
	refer to the same format, even if they aren't the same.

	:param user_ext: (str) User-provided extension for output picture
	:param worker_ext: (str) Extension determined by worker for picture

	:returns: True if extensions refer to the same format, False otherwise
	"""
	extensions = {
		"gif": ["gif"],
		"jpg": ["jpg", "jpeg"],
		"jpeg": ["jpg", "jpeg"],
		"png": ["png"],
		}

	return worker_ext in extensions[user_ext]


def get_image_path(logger, music_path, image_path, worker_ext, forced):
	"""
	Determine the path to store the extracted image to if one was not provided,
	Confirm the user's provided extension matches the one determined to represent
	the picture is a path was provided.

	:param logger: (Logger) Logging object
	:param music_path: (str) Absolute path to music file
	:param image_path: (str/None) Absolute path to store extracted image
	:param worker_ext: (str) Extension worker believes matches the picture
	:param forced: (bool) Whether or not to force using the user's extension,
						  even if it does not match the image format

	:returns: None if there was an error confirming the desired location,
			  String representing absolute path to store image otherwise
	"""
	path = None

	# If the user didn't pass in a path for the extracted image,
	# use the name of the music file with the correct extension
	if image_path is None:
		image_path = ".".join([music_path.rsplit(".", 1)[0], worker_ext])
		logger.debug("User did not provide path for image of file %s", music_path)
		logger.debug("Setting image output path to %s", image_path)

	# Otherwise determine if changing what the user provided is necessary
	else:
		user_ext = image_path.rsplit(".", 1)[1]

		if not fuzzy_match(user_ext, worker_ext):
			logger.debug("Image extension provided by user: %s", user_ext)
			logger.debug("Image extension determined by worker: %s", worker_ext)

			if forced:
				logger.debug("Being forced to use user-provided path")
			else:
				image_path = ".".join([image_path.rsplit(".", 1)[0], worker_ext])
				logger.debug("Changing image output path to %s", image_path)

	# Next, determine if you are required to and allowed to overwrite existing files
	if isfile(image_path):
		logger.info("Image %s already exists", image_path)

		if forced:
			logger.info("Being forced to overwrite")
			path = image_path
		else:
			logger.info("Not overwriting existing file; skipping")
	else:
		path = image_path

	return path


def extract_image(logger, music_path, cover_path, dry_run, forced):
	"""
	Dispatch function handling extracting cover image from music files.

	:param logger: (Logger) Logging object
	:param music_path: (str) Absolute path to music file
	:param cover_path: (str/None) Desired path to store extracted image
	:param dry_run: (bool) Whether or not to actually write images to disk
	:param forced: (bool) Whether or not the tool should do things it doesn't
						  believe are beneficial
	"""
	worker = get_format_worker(music_path)

	if worker is None:
		logger.info("File %s is not a supported music file", music_path)
		return

	(image_data, worker_ext) = worker.get_image_data(logger, music_path)
	cover_path = get_image_path(logger, music_path, cover_path, worker_ext, forced)

	if cover_path is not None:
		logger.debug("Writing image data from %s to %s", music_path, cover_path)
		if not dry_run:
			write_to_disk(logger, cover_path, image_data)



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                           (C) 2015-16, 2019 Mischif                          #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from os.path import join

import pytest

from apic_tool.workers import get_format_worker
from apic_tool.workers.mp3worker import MP3Worker


@pytest.mark.parametrize("filename, worker_type", [("test.mp3", MP3Worker), ("test.xyz", None)],
						 ids=["valid-type", "invalid-type"])
def test_get_format_worker(tmpdir, filename, worker_type):
	full_path = join(str(tmpdir), filename)
	assert get_format_worker(full_path) == worker_type



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                       (C)2015-16, 2019-20 Jeremy Brown                       #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from abc import ABCMeta, abstractmethod

ABC = ABCMeta('ABC', (object,), {})

class BaseWorker(ABC):
	@staticmethod
	@abstractmethod
	def supported_extensions():
		"""
		All the extensions relating to the music this worker can manipulate.

		:returns: (list) Strings representing file extensions
						 this worker can handle
		"""
		raise NotImplementedError("Implement me")

	@staticmethod
	@abstractmethod
	def get_image_data(logger, path):
		"""
		Extract the image data from the given file.

		:param logger: (Logger) Logging object
		:param path: (str) Absolute path to music file

		:returns: (tuple) (None, None) if no image data in file,
						  (bytes, str) Image data from file,
						  			   determined extension of data
		"""
		raise NotImplementedError("Implement me")

	@staticmethod
	@abstractmethod
	def can_insert_image(logger, path, forced):
		"""
		Determine if it is possible to insert an image into the given file.

		:param logger: (Logger) Logging object
		:param path: (str) Absolute path to music file
		:param forced: (bool) Whether issues should be overlooked when trying to insert image

		:returns: (bool) True if it is possible to insert image, False otherwise
		"""
		raise NotImplementedError("Implement me")

	@staticmethod
	@abstractmethod
	def write_to_metadata(logger, music_path, cover_path, forced):
		"""
		Write a given image to a given music file.

		:param logger: (Logger) Logging object
		:param music_path: (str) Absolute path to music file
		:param cover_path: (str) Absolute path to image to write to music file
		:param forced: (bool) Whether issues should be overlooked when trying to insert image
		"""
		raise NotImplementedError("Implement me")



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                           (C) 2015-16, 2019 Mischif                          #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

import pytest

from apic_tool.workers.baseworker import BaseWorker

def test_baseworker():
	with pytest.raises(NotImplementedError):
		BaseWorker.supported_extensions()

	with pytest.raises(NotImplementedError):
		BaseWorker.get_image_data(None, None)

	with pytest.raises(NotImplementedError):
		BaseWorker.can_insert_image(None, None, None)

	with pytest.raises(NotImplementedError):
		BaseWorker.write_to_metadata(None, None, None, None)



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                       (C)2015-16, 2019-20 Jeremy Brown                       #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from apic_tool.workers.mp3worker import MP3Worker

VALID_WORKERS = [MP3Worker]

EXTENSION_WORKER_MAPPING = {
	ext: worker
	for worker in VALID_WORKERS
	for ext in worker.supported_extensions()
	}

SUPPORTED_MUSIC = [ext for worker in VALID_WORKERS for ext in worker.supported_extensions()]

def get_format_worker(path):
	"""
	Get the format worker corresponding to the music file.

	:param path: (str) Absolute path to music file

	:returns: None if no worker for file type,
			  the corresponding format worker otherwise
	"""
	extension = path.rsplit(".", 1)[1]
	return EXTENSION_WORKER_MAPPING.get(extension)



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                           (C) 2015-16, 2019 Mischif                          #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from mock import call, Mock
from os.path import abspath, dirname, getsize, join
from shutil import copy

import pytest

from apic_tool.insertion import get_music_files, insert_image


APIC_TOOL_DATA = abspath(join(dirname(__file__), "data"))
AUTOSORT_AUDIO = abspath(join(dirname(dirname(__file__)), "id3autosort", "audio"))


def test_get_music_files():
	mock_logger = Mock()
	file_list = [join(APIC_TOOL_DATA, "test_insert.mp3")]
	dir_list = [AUTOSORT_AUDIO]
	result = sorted([join(AUTOSORT_AUDIO, "test_mp3.mp3"), join(APIC_TOOL_DATA, "test_insert.mp3")])

	assert sorted(get_music_files(mock_logger, file_list, dir_list, True)) == result
	mock_logger.debug.assert_has_calls([
		call("File %s is not a supported music file, skipping", join(AUTOSORT_AUDIO, "test_aac.m4a")),
		call("File %s is not a supported music file, skipping", join(AUTOSORT_AUDIO, "test_aiff.aiff")),
		call("File %s is not a supported music file, skipping", join(AUTOSORT_AUDIO, "test_flac.flac")),
		call("File %s is not a supported music file, skipping", join(AUTOSORT_AUDIO, "test_ogg.ogg")),
		call("File %s is not a supported music file, skipping", join(AUTOSORT_AUDIO, "test_wav.wav")),
		call("File %s is not a supported music file, skipping", join(AUTOSORT_AUDIO, "test_wma.wma")),
		], any_order=True)


def test_insert_image(tmpdir):
	mock_logger = Mock()
	music_path = join(str(tmpdir), "test_insert.mp3")
	cover_path = join(str(tmpdir), "test_cover.png")

	copy(join(APIC_TOOL_DATA, "test_insert.mp3"), music_path)
	copy(join(APIC_TOOL_DATA, "test_cover.png"), cover_path)
	orig_size = getsize(music_path)
	cover_size = getsize(cover_path)
	assert len(tmpdir.listdir()) == 2

	insert_image(mock_logger, cover_path, None, [music_path], False, False, False)
	assert len(tmpdir.listdir()) == 1
	assert getsize(music_path) >= orig_size + cover_size
	mock_logger.debug.assert_any_call("Writing image %s to file %s", cover_path, music_path)
	mock_logger.info.assert_has_calls([
		call("Deleting image file %s", cover_path)
		], any_order=True)



# encoding: utf-8

################################################################################
#                                 id3autosort                                  #
#  A collection of tools for manipulating and interacting with music metadata  #
#                               (C) 2019 Mischif                               #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import unicode_literals

from errno import EACCES
from os import sep
from os.path import abspath, dirname, join

import pytest

from mock import Mock, patch
from mutagen import File

from id3autosort.sorter import (
	get_music_files,
	get_new_path,
	normalize_tags,
	sort
	)


TEST_AUDIO = abspath(join(dirname(__file__), "audio"))


@pytest.mark.parametrize("audio_path", [
	join(TEST_AUDIO, "test_aac.m4a"),
	join(TEST_AUDIO, "test_aiff.aiff"),
	join(TEST_AUDIO, "test_wma.wma")
	], ids=["easy-tags", "aiff-tags", "wma-tags"])
@pytest.mark.parametrize("windows_safe", [True, False], ids=["windows-safe", "non-windows-safe"])
def test_parse_normalize_tags(audio_path, windows_safe):
	mock_logger = Mock()
	data = File(audio_path, easy=True)

	if "audio/aiff" in data.mime:
		result = {
			"album": "_id3autosort_testing_",
			"date": "2019",
			"genre": "Testing",
			"title": "Test AIFF",
			"tracknumber": "7",
			}

		if windows_safe:
			result.update({"artist": "TestAIFF"})
		else:
			result.update({"artist": ">Test?A\u012cFF<", "album": "_id3autosort:_testing_"})

	elif "audio/x-wma" in data.mime:
		result = {
			"album": "_id3autosort_testing_",
			"genre": "Testing",
			"title": "Test WMA",
			"tracknumber": "4",
			}

		if windows_safe:
			result.update({"artist": "TestWMA"})
		else:
			result.update({"artist": "?Test<>\u0174MA:", "album": "_id3autosort:_testing_"})

	else:
		result = {
			"album": "_id3autosort_testing_",
			"date": "0",
			"genre": "Testing",
			"title": "Test AAC",
			"tracknumber": "3",
			}

		if windows_safe:
			result.update({"artist": "TestAAC"})
		else:
			result.update({"artist": "|Test<>AA\xc7.", "album": "_id3autosort:_testing_"})


	assert result == normalize_tags(mock_logger, data, windows_safe)


@patch("id3autosort.sorter.File")
def test_get_music_files(mock_file):
	mock_logger = Mock()

	def _middle(path, easy):
		if path.endswith(".mp3"):
			raise OSError(EACCES, "Permission Denied")
		else:
			return File(path, easy=easy)

	mock_file.side_effect = _middle

	result = get_music_files(mock_logger, TEST_AUDIO)
	assert len(result) == 5


@patch("id3autosort.sorter.normalize_tags")
def test_get_new_path(mock_normalize, tmpdir):
	mock_logger = Mock()
	tags = {"artist": "Track Artist", "album": "Track Album", "date": "1017"}
	valid_structure = sep.join(["{artist}", "{album} ({date})"])
	invalid_structure = sep.join(["{genre}", "{artist}", "{album}"])

	mock_normalize.return_value = tags

	assert join(str(tmpdir),
				"Track Artist",
				"Track Album (1017)") == get_new_path(mock_logger,
													  str(tmpdir),
													  valid_structure,
													  None,
													  True)
	assert None == get_new_path(mock_logger, str(tmpdir), invalid_structure, None, True)


@patch("id3autosort.sorter.isdir")
@patch("id3autosort.sorter.makedirs")
@patch("id3autosort.sorter.move")
def test_sort(mock_move, mock_makedirs, mock_isdir, tmpdir):
	mock_logger = Mock()

	def _makedirs_middle(path, permissions):
		if path.contains("AIFF"):
			raise OSError(EACCES, "Permission Denied")

	def _move_middle(from_path, to_path):
		if from_path.contains(".mp3"):
			raise OSError(EACCES, "Permission Denied")

	mock_isdir.side_effect = lambda path: False if path.find("AIFF") != -1 else True
	mock_makedirs.side_effect = _makedirs_middle
	mock_move.side_effect = _move_middle

	sort(mock_logger, TEST_AUDIO, "/tmp", "{artist}/{album} ({date})", True, False)
	sort(mock_logger, str(tmpdir), "/tmp", "{artist}/{album} ({date})", True, False)

	assert mock_makedirs.call_count == 3
	assert mock_move.call_count == 2



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                       (C)2015-16, 2019-20 Jeremy Brown                       #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from argparse import Action, ArgumentParser, ArgumentTypeError
from logging import (
	DEBUG,
	ERROR,
	Formatter,
	getLogger,
	INFO,
	StreamHandler,
	WARNING,
	)
from os import access, walk, R_OK, W_OK
from os.path import abspath, dirname, expanduser, isdir, isfile
from sys import argv

from apic_tool import __version__, SUPPORTED_IMAGES, SUPPORTED_MUSIC
from apic_tool.extraction import extract_image
from apic_tool.insertion import insert_image


logger = getLogger(__file__)

class CustomLogs(Formatter):
	FORMATS = {
		ERROR:		"[-] %(message)s",
		WARNING:	"[~] %(message)s",
		INFO:		"[*] %(message)s",
		DEBUG:		"[.] %(message)s",
		}

	def __init__(self):
		super(CustomLogs, self).__init__()

	def format(self, record):
		log_style = getattr(self, "_style", self)
		log_style._fmt = self.FORMATS.get(record.levelno, self.FORMATS[DEBUG])
		return super(CustomLogs, self).format(record)


class AbsoluteAccessiblePaths(Action):
	def __call__(self, parser, namespace, values, option_string=None):
		if values != None:
			out = getattr(namespace, self.dest, [])
			if out is None:
				out = []

			# Remember each item in an option list is passed through the action individually
			if isinstance(values, list):
				values = values[0]
			expanded_path = abspath(expanduser(values))

			# Confirm file paths point to files and directory paths point to directories
			# Since the path pointed to by extract_pic may not currently exist, don't raise
			# if it doesn't
			if self.dest == "extract_pic":
				pass
			elif self.dest in ["insert_files", "insert_pic", "extract_music"]:
				if not isfile(expanded_path):
					raise ArgumentTypeError("The given path is not a file: {0}".format(expanded_path))
			else:
				if not isdir(expanded_path):
					raise ArgumentTypeError("The given path is not a directory: {0}".format(expanded_path))

			# Confirm this is a supported image
			if self.dest == "insert_pic":
				if expanded_path.rsplit(".", 1)[1].lower() not in SUPPORTED_IMAGES:
					raise ArgumentTypeError("Unsupported image type: {0}".format(expanded_path))

			# Confirm this is a supported music file
			if self.dest == "extract_music":
				if expanded_path.rsplit(".", 1)[1].lower() not in SUPPORTED_MUSIC:
					raise ArgumentTypeError("Unsupported music type: {0}".format(expanded_path))


			if self.dest in ["insert_dirs", "insert_files"]:
				out.append(expanded_path)
			else:
				out = expanded_path

			setattr(namespace, self.dest, out)


def parse_args(**kwargs):
	"""
	Read arguments from stdin while validating and performing any necessary conversions

	:returns: (Namespace) Tool arguments
	"""
	main_parser = ArgumentParser(
		prog = "apic-tool",
		description = "Inserts and extracts cover images to/from music files.",
		epilog = "(C) 2015-16, 2019-20 Jeremy Brown; released under Non-Profit Open Source License version 3.0"
		)

	main_parser.add_argument("-n", "--dry-run",
							 action="store_true",
							 help="Simulate the actions instead of actually doing them"
							 )

	main_parser.add_argument("-v", "--verbose",
							 action="store_true",
							 help="Increase logging verbosity")

	main_parser.add_argument("--force",
							 action="store_true",
							 dest="force",
							 help="Whether or not the tool should allow things to happen that may have complications")

	main_parser.add_argument("--version",
							 action="version",
							 version="%(prog)s {}".format(__version__))


	tool_actions = main_parser.add_subparsers(title="Actions", metavar="")
	extract_parser = tool_actions.add_parser("extract", help="Extract image from a music file")
	extract_parser.set_defaults(action="extract")
	insert_parser = tool_actions.add_parser("insert", help="Insert image into a music file")
	insert_parser.set_defaults(action="insert")
	insert_arg = insert_parser.add_mutually_exclusive_group(required=True)

	insert_arg.add_argument("-d", "--dir",
							action=AbsoluteAccessiblePaths,
							dest="insert_dirs",
							metavar="DIR",
							nargs="+",
							help="Director(y|ies) containing files to manipulate"
							)

	insert_arg.add_argument("-f", "--file",
							action=AbsoluteAccessiblePaths,
							dest="insert_files",
							nargs="+",
							help="Input file to manipulate"
							)

	insert_parser.add_argument("-p", "--pic",
							   action=AbsoluteAccessiblePaths,
							   dest="insert_pic",
							   help="Image to insert",
							   required=True
							   )

	insert_parser.add_argument("-k", "--keep",
							   action="store_true",
							   dest="keep_pic",
							   help="Don't delete image after inserting it"
							   )

	extract_parser.add_argument("extract_music",
								action=AbsoluteAccessiblePaths,
								help="File to extract image from"
								)

	extract_parser.add_argument("extract_pic",
								action=AbsoluteAccessiblePaths,
								default=None,
								nargs="?",
								help="Filename to send extracted image to"
								)

	return main_parser.parse_args(kwargs.get("argv", argv[1:]))


def main():
	"""
	Tool entry point
	"""
	args = parse_args()

	logger.setLevel(DEBUG if args.verbose else INFO)
	log_hdlr = StreamHandler()
	log_hdlr.setFormatter(CustomLogs())
	logger.addHandler(log_hdlr)

	logger.debug("Dry run: %s", args.dry_run)
	logger.debug("Forcing: %s", args.force)

	if args.action == "extract":
		logger.debug("Extraction file: %s", args.extract_music)
		logger.debug("Extraction result: %s", args.extract_pic)
		extract_image(logger, args.extract_music, args.extract_pic, args.dry_run, args.force)
	else:
		logger.debug("Insertion files: %s", args.insert_files)
		logger.debug("Insertion directories: %s", args.insert_dirs)
		logger.debug("Cover to insert: %s", args.insert_pic)
		logger.debug("Keep covers after insertion: %s", args.keep_pic)
		insert_image(logger, args.insert_pic, args.insert_dirs, args.insert_files, args.keep_pic, args.dry_run, args.force)



# encoding: utf-8

################################################################################
#                                 id3autosort                                  #
#                      Sort audio files based on metadata                      #
#                    (C)2009-10, 2015, 2019-20 Jeremy Brown                    #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

import re

from os import makedirs, walk
from os.path import isdir, join
from shutil import move
from unicodedata import normalize

from mutagen import File


PATH_CHARS = re.compile("[/\\\\]")
WINDOWS_UNSAFE_CHARS = re.compile("[:<>\"\*\?\|]")


def normalize_tags(logger, md, windows_safe):
	"""
	Modify or remove characters in tags that would cause issues when stored on a filesystem.
	Additionally, handle formats Mutagen recognizes but does not have an easy-mode version for,
	creating one for use by downstream consumers.

	:param logger: (Logger) Logging object
	:param md: Mutagen metadata structure
	:param windows_safe: (bool) Whether or not to perform extra normalization for Windows platforms

	:returns: (dict) A normalized dict of tags
	"""
	normalized = {}

	def _structure_aiff_tags(tags):
		structured = {}

		mapping = {
			"TALB": "album",
			"TCON": "genre",
			"TDRC": "date",
			"TIT2": "title",
			"TPE1": "artist",
			"TRCK": "tracknumber",
			}

		for m in filter(lambda t: t in tags, mapping.keys()):
			structured[mapping[m]] = tags[m].text

		return structured

	def _structure_wma_tags(tags):
		structured = {}

		mapping = {
			"Author": "artist",
			"Title": "title",
			"WM/AlbumTitle": "album",
			"WM/Genre": "genre",
			"WM/TrackNumber": "tracknumber",
			"year": "date",
			}

		for m in filter(lambda t: t in tags, mapping.keys()):
			structured[mapping[m]] = tags[m][0].value

		return structured

	# Mutagen doesn't have easy mode for all audio formats,
	# convert those certain formats' tags to easy mode manually
	unstructured_mimes = {"audio/aiff": _structure_aiff_tags, "audio/x-wma": _structure_wma_tags}
	unstructured = list(filter(lambda m: m in md.mime, unstructured_mimes.keys()))
	if unstructured:
		raw_tags = unstructured_mimes[unstructured[0]](md.tags)
	else:
		raw_tags = md.tags


	logger.debug("Original tags: %s", raw_tags)

	# Since there can be multiple entries per tag,
	# tag values are lists instead of a simple string;
	# Just use the first value even if it is a multi-item tag
	# Encode/decode combo because date value might not be a string
	for (k, v) in raw_tags.items():
		if isinstance(v, list):
			this_value = v[0].encode("utf-8").decode("utf-8")
		else:
			this_value = v.encode("utf-8").decode("utf-8")

		# Skip empty tags
		if not this_value:
			logger.debug("Skipping empty tag %s", k)
			continue

		# Explicitly replace backslashes and forward slashes with underscores
		# regardless of platform
		this_value = PATH_CHARS.sub("_", this_value)

		if windows_safe:
			# Remove the other characters Windows doesn't like
			this_value = WINDOWS_UNSAFE_CHARS.sub("", this_value)

			# Convert characters the furriners use to good-ol' 'MURICAN letters
			# NTFS is probably fine, but FAT32 ruins everything good in this world
			this_value = normalize("NFD", this_value).encode("ascii", "ignore").decode("ascii")

			# Unix-based OSs are fine with an NTFS directory ending with a period,
			# but Windows will refuse to open them,
			# so make sure directories don't end in a period for Windows-safety
			if this_value[-1] == ".":
				this_value = this_value[:-1]

		normalized[k] = this_value

	logger.debug("Normalized tags: %s", normalized)
	return normalized


def get_music_files(logger, music_dir):
	"""
	Obtain a list of all music files Mutagen can read metadata for inside the given directory.

	:param logger: (logger) Logging object
	:param music_dir: (str) Absolute path to directory to check for music files

	:returns: (list) All music files Mutagen can read metadata for inside the given directory
	"""
	valid_files = []

	for (basedir, dirs, basenames) in walk(music_dir):
		maybe_valid = [join(basedir, name) for name in basenames]

		for path in maybe_valid:
			try:
				logger.debug("Attempting to parse %s", path)
				parsed_file = File(path, easy=True)
			except Exception as e:
				logger.info("Exception attempting to read file %s: %s", path, e)
			else:
				if parsed_file is None:
					logger.debug("File %s has no music metadata", path)
				else:
					valid_files.append((path, parsed_file))

	return valid_files


def get_new_path(logger, out_dir, structure, metadata, windows_safe):
	"""
	Determine the new location the given file should be located,
	based upon the file's metadata and the provided folder structure.

	:param logger: (Logger) Logging object
	:param out_dir: (str) Root directory the file should be moved to
	:param structure: (str) Desired structure for music files inside root directory
	:param metadata: Mutagen metadata structure for given file
	:param windows_safe: (bool) Whether or not to perform extra normalization for Windows platforms

	:returns: None if there was an error determining the new location,
			  a string representing an absolute path otherwise.
	"""
	new_path = None
	tags = normalize_tags(logger, metadata, windows_safe)

	try:
		new_path = join(out_dir, structure.format(**tags))
	except Exception as e:
		logger.debug("Error making new path: %s", e)
	else:
		logger.debug("Subdirectory structure for file: %s", new_path)

	return new_path


def sort(logger, in_dir, out_dir, structure, windows_safe, dry_run):
	"""
	Main function handling finding music, finding the location said music
	should be moved to, and moving it.

	:param logger: (Logger) Logging object
	:param in_dir: (str) Absolute path to music source directory
	:param out_dir: (str) Absolute path to music destination directory
	:param windows_safe: (bool) Whether or not to perform extra normalization for Windows platforms
	:param dry_run: (bool) Whether or not to perform actual movement of files
	"""
	files_with_metadata = get_music_files(logger, in_dir)

	if not len(files_with_metadata):
		logger.info("No music files in %s", in_dir)

	for (file_path, metadata) in files_with_metadata:
		new_path = get_new_path(logger, out_dir, structure, metadata, windows_safe)

		if new_path is None:
			logger.info("File %s does not have tags to fulfill specified structure, skipping",
					 file_path)
			continue
		else:
			logger.debug("Moving file %s to %s", file_path, new_path)

			if not dry_run:
				try:
					makedirs(new_path, 0o755)
				except Exception as e:
					# It's fine if the directory already exists
					if getattr(e, "errno", None) != 17:
						logger.info("Could not create destination folders for file %s: %s",
								 file_path, e)

				if isdir(new_path):
					try:
						move(file_path, new_path)
					except Exception as e:
						logger.info("Could not move file %s to new location: %s", file_path, e)



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                           (C) 2015-16, 2019 Mischif                          #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from errno import EACCES
from functools import partial
from os.path import abspath, dirname, join
from mock import call, Mock, patch

try:
	from importlib import reload
except ImportError:
	pass

import pytest

from mutagen import MutagenError
from mutagen.id3 import ID3, APIC
from mutagen.mp3 import MP3

from apic_tool.workers import mp3worker


APIC_TOOL_DATA = abspath(join(dirname(dirname(__file__)), "data"))
AUTOSORT_AUDIO = abspath(join(dirname(dirname(dirname(__file__))), "id3autosort", "audio"))


def test_supported_extensions():
	assert mp3worker.MP3Worker.supported_extensions() == ["mp3"]


@patch("apic_tool.workers.mp3worker.MP3")
@pytest.mark.parametrize("scenario", ["good", "eacces", "exception"],
						 ids=["happy-path", "permission-denied", "generic-exception"])
def test_load_file(mock_mp3, scenario):
	mock_logger = Mock()

	def _middle(path):
		raise MutagenError(IOError(EACCES, "Permission denied"))

	if scenario == "eacces":
		mock_mp3.side_effect = _middle
	else:
		mock_mp3.side_effect = MP3

	if scenario == "exception":
		path = join(AUTOSORT_AUDIO, "test_wav.wav")
	else:
		path = join(APIC_TOOL_DATA, "test_extract.mp3")

	if scenario == "good":
		assert isinstance(mp3worker.MP3Worker.load_file(mock_logger, path), MP3)
	elif scenario == "eacces":
		assert mp3worker.MP3Worker.load_file(mock_logger, path) is None
		mock_logger.info.assert_called_once_with("Permission denied attempting to access file %s", path)
	elif scenario == "exception":
		assert mp3worker.MP3Worker.load_file(mock_logger, path) is None
		mock_logger.info.assert_called_once_with("Error trying to load %s as MP3 file: %s", path, "can't sync to MPEG frame")


@pytest.mark.parametrize("scenario", ["good", "missing", "unclean", "tagless"],
						 ids=["happy-path", "missing-metadata", "sketchy-load", "tagless-file"])
def test_get_image_data(scenario):
	mp3worker.MP3Worker.load_file = mock_load = Mock()
	mock_logger = Mock()
	path = join(APIC_TOOL_DATA, "test_extract.mp3")

	with open(join(APIC_TOOL_DATA, "test_cover.png"), "rb") as f:
		expected_cover = f.read()

	if scenario == "good":
		reload(mp3worker)
	elif scenario == "missing":
		mock_load.return_value = None
	else:
		mock_load.return_value = mock_load

	if scenario == "unclean":
		mock_load.info.sketchy = True
		mock_load.tags = {}
	elif scenario == "tagless":
		mock_load.info.sketchy = False
		mock_load.tags = {}

	(data, ext) = mp3worker.MP3Worker.get_image_data(mock_logger, path)

	assert (data, ext) == (expected_cover, "png") if scenario == "good" else (None, None)

	if scenario == "unclean":
		mock_logger.warning.assert_called_once_with("Couldn't load file %s cleanly", path)
		mock_logger.info.assert_called_once_with("No tags in file %s, skipping", path)
	elif scenario == "tagless":
		mock_logger.info.assert_called_once_with("No tags in file %s, skipping", path)


@pytest.mark.parametrize("scenario", ["good", "missing", "unclean", "tagless", "forced"],
						 ids=["happy-path", "missing-metadata", "sketchy-load", "tagless-file", "forced"])
def test_can_insert_image(scenario):
	mp3worker.MP3Worker.load_file = mock_load = Mock()
	mock_logger = Mock()
	path = join(APIC_TOOL_DATA, "test_extract.mp3")
	force = scenario == "forced"

	if scenario == "good":
		reload(mp3worker)
	elif scenario == "missing" or scenario == "forced":
		mock_load.return_value = None
	else:
		mock_load.return_value = mock_load

	if scenario == "unclean":
		mock_load.info.sketchy = True
	elif scenario == "tagless":
		mock_load.info.sketchy = False
		mock_load.tags = {}

	result = mp3worker.MP3Worker.can_insert_image(mock_logger, path, force)

	assert result is (True if scenario == "good" or force else False)

	if scenario == "unclean":
		mock_logger.warning.assert_called_once_with("Couldn't load file %s cleanly, skipping", path)
	elif scenario == "tagless":
		mock_logger.info.assert_called_once_with("No tags in file %s, skipping", path)


@pytest.mark.parametrize("scenario", ["good", "bad"], ids=["happy-path", "unhappy-path"])
def test_write_to_metadata(scenario):
	mp3worker.MP3Worker.load_file = mock_load = Mock()
	mock_logger = Mock()
	music_path = join(APIC_TOOL_DATA, "test_extract.mp3")
	cover_path = join(APIC_TOOL_DATA, "test_cover.png")

	if scenario == "good":
		reload(mp3worker)
	elif scenario == "bad":
		mock_load.return_value = mock_load

	def _add_tags(mock):
		mock.tags = ID3()
		mock.tags.version = (2, 2, 7)
		mock.tags.save = Mock()
		mock.tags.save.side_effect = Exception("Test error")
		tag = APIC(
			   encoding=3,			# UTF-8
			   type=3,				# Cover image
			   mime="png",
			   )
		mock.tags.add(tag)

	mock_load.info.sketchy = True
	mock_load.tags = {}
	mock_load.add_tags.side_effect = partial(_add_tags, mock_load)

	if scenario == "good":
		assert mp3worker.MP3Worker.write_to_metadata(mock_logger, music_path, cover_path, False) is True
		mock_logger.info.assert_called_once_with("File %s already has embedded image, skipping", music_path)
	elif scenario == "bad":
		assert mp3worker.MP3Worker.write_to_metadata(mock_logger, music_path, cover_path, True) is False
		mock_logger.debug.assert_has_calls([
			call("Forced to continue with file %s despite unclean load", music_path),
			call("Forced to create metadata for file %s, which has none", music_path),
			call("Tags are version %d.%d", 2, 2),
			call("Upgrading tags for file %s to v2.3", music_path),
			call("Supposed mimetype for image: %s", "image/png"),
			call("Adding image to file")
			])
		mock_logger.info.assert_has_calls([
			call("Forced to remove pre-existing APIC tags for file %s", music_path),
			call('Saving updated tags'),
			call("Error saving tags for file %s: %s", music_path, "Test error")
			])


# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                       (C)2015-16, 2019-20 Jeremy Brown                       #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from errno import EACCES
from imghdr import what
from mimetypes import guess_type

import builtins

from mutagen import MutagenError
from mutagen.id3 import ID3, APIC
from mutagen.mp3 import MP3

from apic_tool.workers.baseworker import BaseWorker


SUPPORTED_EXTENSIONS = ["mp3"]


class MP3Worker(BaseWorker):
	@staticmethod
	def supported_extensions():
		"""
		All the extensions relating to the music this worker can manipulate.

		:returns: (list) Strings representing file extensions
						 this worker can handle
		"""
		return SUPPORTED_EXTENSIONS

	@staticmethod
	def load_file(logger, path):
		"""
		Obtain ID3 data for the given file.

		:param logger: (Logger) Logging object
		:param path: (str) Absolute path to MP3 file

		:returns: (MP3/None) None if there was a major issue loading metadata,
							 MP3 object otherwise
		"""
		music = None

		try:
			music = MP3(path)
		except MutagenError as e:
			orig_e = e.args[0]
			if isinstance(orig_e, getattr(builtins, "PermissionError", IOError)) and orig_e.errno == EACCES:
				logger.info("Permission denied attempting to access file %s", path)
			else:
				logger.info("Error trying to load %s as MP3 file: %s", path, str(orig_e))

		return music

	@staticmethod
	def get_image_data(logger, path):
		"""
		Extract the image data from the given file.

		:param logger: (Logger) Logging object
		:param path: (str) Absolute path to music file

		:returns: (tuple) (None, None) if no image data in file,
						  (bytes, str) Image data from file,
						  			   determined extension of data
		"""
		data = None
		ext = None
		music = MP3Worker.load_file(logger, path)

		if music is not None:
			if music.info.sketchy:
				logger.warning("Couldn't load file %s cleanly", path)

			if not music.tags:
				logger.info("No tags in file %s, skipping", path)
			else:
				images = music.tags.getall("APIC")

				if images:
					data = images[0].data
					ext = what(path, data)

		return (data, ext)

	@staticmethod
	def can_insert_image(logger, path, forced):
		"""
		Determine if it is possible to insert an image into the given file.

		:param logger: (Logger) Logging object
		:param path: (str) Absolute path to music file
		:param forced: (bool) Whether issues should be overlooked when trying to insert image

		:returns: (bool) True if it is possible to insert image, False otherwise
		"""
		result = forced
		music = MP3Worker.load_file(logger, path)

		if music is not None and not forced:
			if music.info.sketchy:
				logger.warning("Couldn't load file %s cleanly, skipping", path)

			elif not music.tags:
				logger.info("No tags in file %s, skipping", path)

			else:
				result = True

		return result

	@staticmethod
	def write_to_metadata(logger, music_path, cover_path, forced):
		"""
		Write a given image to a given music file.

		:param logger: (Logger) Logging object
		:param music_path: (str) Absolute path to music file
		:param cover_path: (str) Absolute path to image to write to music file
		:param forced: (bool) Whether issues should be overlooked when trying to insert image
		"""
		result = False
		music = MP3Worker.load_file(logger, music_path)

		if music is not None:
			if music.info.sketchy:
				if not forced:
					return result
				else:
					logger.debug("Forced to continue with file %s despite unclean load", music_path)

			if not music.tags:
				if not forced:
					return result
				else:
					logger.debug("Forced to create metadata for file %s, which has none", music_path)
					music.add_tags()

			images = music.tags.getall("APIC")

			if images and not forced:
				logger.info("File %s already has embedded image, skipping", music_path)
				result = True

			else:
				if images:
					logger.info("Forced to remove pre-existing APIC tags for file %s", music_path)
					music.tags.delall("APIC")

				tag_version = (music.tags.version[0], music.tags.version[1])
				old_tags = True if tag_version < (2, 3) else False
				logger.debug("Tags are version %d.%d", *tag_version)

				if old_tags:
					logger.debug("Upgrading tags for file %s to v2.3", music_path)
					music.tags.update_to_v23()

				mimetype = guess_type(cover_path)[0]
				logger.debug("Supposed mimetype for image: %s", mimetype)

				with open(cover_path, "rb") as cover:
					tag = APIC(
							   encoding=3,			# UTF-8
							   type=3,				# Cover image
							   mime=mimetype,
							   data=cover.read(),
							   )

					logger.debug("Adding image to file")
					music.tags.add(tag)

				logger.info("Saving updated tags")
				try:
					music.tags.save(music_path, v2_version=3 if old_tags else tag_version[1])
				except Exception as e:
					logger.info("Error saving tags for file %s: %s", music_path, str(e))
				else:
					result = True

			return result



# encoding: utf-8

################################################################################
#                                  apic-tool                                   #
#       Insert cover images to and extract cover images from music files       #
#                           (C) 2015-16, 2019 Mischif                          #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from errno import EACCES
from mock import Mock, patch
from os import name, sep
from os.path import abspath, dirname, exists, join

import pytest

from apic_tool.extraction import extract_image, fuzzy_match, get_image_path, write_to_disk


TEST_DATA = abspath(join(dirname(__file__), "data"))


@pytest.mark.parametrize("scenario", ["eacces", "ioerror", "good"],
						 ids=["permission-denied", "generic-ioerror", "good"])
def test_write_to_disk(tmpdir, scenario):
	mock_logger = Mock()

	if scenario == "good":
		out_path = join(str(tmpdir), "test_write_to_disk.png")
	elif scenario == "eacces":
		if name == "posix":
			out_path = join(sep, "sys", "test_write_to_disk.png")
		else:
			out_path = join("C:", sep, "Windows", "explorer.exe")
	elif scenario == "ioerror":
		out_path = join("/proc", "test_write_to_disk.png")

	with open(join(TEST_DATA, "test_cover.png"), "rb") as f:
		data = f.read()

	try:
		write_to_disk(mock_logger, out_path, data)
	except IOError as e:
		assert e.errno != EACCES
		assert scenario == "ioerror"
	else:
		if scenario == "good":
			mock_logger.info.assert_not_called()
		elif scenario == "eacces":
			mock_logger.info.assert_called_once_with("Permission denied attempting to write image %s", out_path) 


@pytest.mark.parametrize("user", ["gif", "jpg", "jpeg", "png"],
						 ids=["user-gif", "user-jpg", "user-jpeg", "user-png"])
@pytest.mark.parametrize("worker", ["gif", "jpg", "jpeg", "png"],
						 ids=["worker-gif", "worker-jpg", "worker-jpeg", "worker-png"])
def test_fuzzy_match(user, worker):
	if user == worker:
		result = True
	elif user == "jpg" and worker == "jpeg":
		result = True
	elif user == "jpeg" and worker == "jpg":
		result = True
	else:
		result = False

	assert fuzzy_match(user, worker) is result


@pytest.mark.parametrize("image_name", [None, "test_file.jpg", "test_file.png"],
						 ids=["no-path", "correct-ext", "incorrect-ext"])
@pytest.mark.parametrize("forced", [True, False], ids=["forced", "not-forced"])
@pytest.mark.parametrize("exists", [True, False], ids=["file-exists", "new-file"])
def test_get_image_path(tmpdir, image_name, forced, exists):
	music_path = join(str(tmpdir), "test_file.mp3")
	mock_logger = Mock()
	result = None
	if image_name is not None:
		user_path = join(str(tmpdir), image_name)
		real_path = user_path if forced else join(str(tmpdir), "test_file.jpg")
	else:
		user_path = None
		real_path = join(str(tmpdir), "test_file.jpg")

	if exists:
		open(real_path, "a").close()

	if (not exists) or (exists and forced):
		result = real_path

	assert get_image_path(mock_logger, music_path, user_path, "jpg", forced) == result


@pytest.mark.parametrize("supported_format", [True, False], ids=["supported-format", "unsupported-format"])
@patch("apic_tool.extraction.write_to_disk")
def test_extract_image(mock_write, supported_format, tmpdir):
	mock_logger = Mock()
	audio_path = join(TEST_DATA, "test_extract.mp3" if supported_format else "test_extract.xyz")
	cover_location = join(str(tmpdir), "test_extract.png")

	with open(join(TEST_DATA, "test_cover.png"), "rb") as f:
		expected_cover = f.read()

	extract_image(mock_logger, audio_path, cover_location, False, False)

	if supported_format:
		mock_write.assert_called_once_with(mock_logger, cover_location, expected_cover)
	else:
		mock_write.assert_not_called()
		mock_logger.info.assert_called_once_with("File %s is not a supported music file", audio_path)



# encoding: utf-8

################################################################################
#                             music-metadata-tools                             #
#  A collection of tools for manipulating and interacting with music metadata  #
#                  (C) 2009-10, 2015-16, 2019-20 Jeremy Brown                  #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from io import open
from os import path
from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as desc:
	long_description = desc.read()


setup(
	name="music-metadata-tools",

	packages=find_packages(where="src"),

	package_dir={"": "src"},

	license="NPOSL-3.0",

	url="https://github.com/mischif/music-metadata-tools",

	description="A collection of tools for manipulating and interacting with music metadata.",

	long_description=long_description,
	long_description_content_type="text/markdown",

	author="Jeremy Brown",
	author_email="mischif@users.noreply.github.com",

	python_requires="~=3.6",

	install_requires=["mutagen"],

	setup_requires=["pytest-runner", "setuptools_scm"],

	tests_require=["mock", "pytest", "pytest-cov"],

	zip_safe=False,

	keywords=["ID3", "APIC"],

	extras_require={
		"test": ["codecov"],
		},

	entry_points={
		"console_scripts": [
							"id3autosort=id3autosort.cli:main",
							"apic-tool=apic_tool.cli:main"
						   ],
		},

	options={
		"aliases": {
			"test": "pytest",
			},

		"metadata": {
			"license_files": "LICENSE",
			},
		},

	classifiers=[
		"Development Status :: 5 - Production/Stable",

		"Operating System :: OS Independent",

		"License :: OSI Approved :: Open Software License 3.0 (OSL-3.0)",

		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		],
	)



# encoding: utf-8

################################################################################
#                                 id3autosort                                  #
#  A collection of tools for manipulating and interacting with music metadata  #
#                               (C) 2019 Mischif                               #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import unicode_literals

from argparse import Namespace
from os import sep
from os.path import abspath, dirname, join

import pytest

from mock import Mock, patch

from id3autosort.cli import main, parse_args


TEST_AUDIO = abspath(join(dirname(__file__), "audio"))


@pytest.mark.parametrize("source_dir", [TEST_AUDIO, join(TEST_AUDIO, "test_mp3.mp3"), "/proc/noexist"],
						 ids=["valid-source", "source-file", "source-inaccessible"])
@pytest.mark.parametrize("windows_safe, verbose, dry_run, structure", [
						 [True, False, False, False],
						 [False, True, False, False],
						 [False, False, True, False],
						 [False, False, False, True]],
						 ids=["windows-safe", "verbose", "dry-run", "custom-structure"])
def test_parse_args(tmpdir, windows_safe, verbose, dry_run, structure, source_dir):
	args_list = []
	output_structure = sep.join(["{artist}", "{album}"])

	if verbose:
		args_list.append("-v")

	if dry_run:
		args_list.append("-n")

	if structure:
		output_structure = sep.join(["{artist}", "{album} ({date})"])
		args_list.extend(["-s", sep.join(["r", "l (d)"])])

	if not windows_safe:
		args_list.append("-u")

	args_list.extend([source_dir, str(tmpdir)])

	if source_dir != TEST_AUDIO:
		with pytest.raises(SystemExit):
			parse_args(argv=args_list)
	else:
		args = parse_args(argv=args_list)

		assert args.verbose == verbose
		assert args.dry_run == dry_run
		assert args.windows_safe == windows_safe
		assert args.structure == output_structure
		assert args.src_paths == [TEST_AUDIO]
		assert args.dest_path == str(tmpdir)


@patch("id3autosort.cli.sort")
@patch("id3autosort.cli.parse_args")
@patch("id3autosort.cli.logger")
def test_main(mock_logger, mock_parse_args, mock_sort):
	args_dict = {
		"dest_path": "/tmp",
		"dry_run": False,
		"src_paths": [TEST_AUDIO],
		"structure": "{artist}/{album}",
		"verbose": False,
		"windows_safe": True,
		}

	mock_parse_args.return_value = Namespace(**args_dict)
	main()

	mock_sort.assert_called_once_with(mock_logger,
									  args_dict["src_paths"][0],
									  args_dict["dest_path"],
									  args_dict["structure"],
									  args_dict["windows_safe"],
									  args_dict["dry_run"])
